/*
 *
 * apps/videoClient.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/core/ReactiveEngine.H>

#include <nucleo/utils/AppUtils.H>
#include <nucleo/utils/SignalUtils.H>

#include <nucleo/image/source/ImageSource.H>

#include <nucleo/image/processing/convolution/Convolution.H>
#include <nucleo/image/processing/convolution/Blur.H>
#include <nucleo/image/processing/difference/Difference.H>
#include <nucleo/image/processing/basic/Resize.H>
#include <nucleo/image/processing/basic/Transform.H>

#include <nucleo/image/sink/ImageSink.H>

#include <string>

#include <stdexcept>

#include <fstream>

using namespace nucleo ;

bool signaled=false ;
void exitProc(int s) {
  std::cerr << "signal : " << s << std::endl ;
  signaled = true ; 
}

int
main(int argc, char **argv) {
  try {
    char *SOURCE = 0 ;
    std::string FILTER = "None" ;
    char *SINK1=0, *SINK2=0 ;
    bool WAIT = false ;
    bool RESTART = false ;
    int VERBOSITY = 0 ;
    char *ENCODING = "PREFERRED" ;
    bool FORCE_SELECT_ENGINE = false ;

    if (parseCommandLine(argc, argv, "v:e:i:f:o:p:wrs", "issSssbbb",
					&VERBOSITY,
					&ENCODING,
					&SOURCE, &FILTER, &SINK1, &SINK2,
					&WAIT, &RESTART,
					&FORCE_SELECT_ENGINE)<0) {
	 std::cerr << std::endl << argv[0] << " [-v verbosity] [-w(ait)] [-i source] [-e encoding] [-f filter] [-o sink1] [-p sink2] [-r(estart source when it stops)]" ;
	 std::cerr << std::endl ;
	 exit(1) ;
    }

    if (!SOURCE) SOURCE = getenv("NUCLEO_SRC1") ;
    if (!SOURCE) SOURCE = "videoin:/anydev/anynode" ;

    if (!SINK1) SINK1 = getenv("NUCLEO_DST1") ;
    if (!SINK1) SINK1 = "glwindow:" ;

    if (FORCE_SELECT_ENGINE)
	 ReactiveEngine::useNativeEngine(false) ;

    Image::Encoding encoding = Image::getEncodingByName(ENCODING) ;
    if (VERBOSITY) std::cerr << "encoding is " << ENCODING << std::endl ;
    ImageSource *source = ImageSource::create(SOURCE, encoding) ;

    Image img ;

    ImageFilter *filter=0 ;
    if (FILTER=="edges")
	 filter = (ImageFilter *) new Convolution_3x3(0,-1,0, -1,4,-1,0, -1,0, 0,1) ;
    else if (FILTER=="horizontals")
	 filter = (ImageFilter *) new Convolution_3x3(-1,-1,-1, 0,0,0, 1,1,1, 1) ;
    else if (FILTER=="verticals")
	 filter = (ImageFilter *) new Convolution_3x3(-1,0,1, -1,0,1, -1,0,1, 1) ;
    else if (FILTER=="blur")
	 filter = (ImageFilter *) new BlurFilter(BlurFilter::HandV, 20, 1) ;
    else if (FILTER=="blur3")
	 filter = (ImageFilter *) new BlurFilter(BlurFilter::HandV, 20, 3) ;
    else if (FILTER=="boost")
	 filter = (ImageFilter *) new Convolution_3x3(-1,-1,-1, -1,21.6,-1, -1,-1,-1, 0,9) ;
    else if (FILTER=="emboss")
	 filter = (ImageFilter *) new Convolution_3x3(-1,-1,0, -1,0,1, 0,1,1, 120,1) ;
    else if (FILTER=="difference")
	 filter = (ImageFilter *) new DifferenceFilter ;
    else if (FILTER=="nobackground")
	 filter = (ImageFilter *) new DifferenceFilter(25,1) ;
    else if (!strncmp(FILTER.c_str(),"resize-",7))
	 filter = (ImageFilter *) new ResizeFilter(FILTER.c_str()+7) ;
    else if (!strncmp(FILTER.c_str(),"crop-",5))
	 filter = (ImageFilter *) new CropFilter(FILTER.c_str()+5) ;
    else if (FILTER=="flip-h")
	 filter = (ImageFilter *) new mirrorFilter('h') ;
    else if (FILTER=="flip-v")
	 filter = (ImageFilter *) new mirrorFilter('v') ;
    else if (FILTER=="rot+90")
	 filter = (ImageFilter *) new rotateFilter(true) ;
    else if (FILTER=="rot-90")
	 filter = (ImageFilter *) new rotateFilter(false) ;

    ImageSink *sink1 = ImageSink::create(SINK1) ;
 
    ImageSink *sink2=0 ;
    if (SINK2) sink2 = ImageSink::create(SINK2) ;

    trapAllSignals(exitProc) ;

    source->start() ;
    sink1->start() ;
    if (SINK2) sink2->start() ;

    while (!signaled && sink1->isActive()) {

	 if (!source->isActive()) {
	   if (RESTART) {
		if (VERBOSITY>1) std::cerr << "R" << std::flush ;
		source->start() ;
	   }
	   else if (!WAIT) break ;
	 }

	 ReactiveEngine::step() ;

	 if (source->getNextImage(&img)) {
	   if (VERBOSITY>1) std::cerr << "." << std::flush ;
	   if (filter) filter->filter(&img) ;
	   sink1->handle(&img) ;
	   if (SINK2) sink2->handle(&img) ;
	 }
    }

    if (VERBOSITY) {
	 std::cerr << std::endl ;
	 std::cerr << "source: " 
			 << source->getFrameNumber() << " frame(s), "
			 << source->getFrameRate() << " fps" << std::endl ;
	 std::cerr << "sink1: "
			 << sink1->getFrameNumber() << " frame(s), "
			 << sink1->getFrameRate() << " fps" << std::endl ;
	 if (sink2) std::cerr << "sink2: " << sink2->getFrameRate() << " fps" << std::endl ;
    }

    delete source ;
    delete sink1 ;
    delete sink2 ;

  } catch (std::runtime_error e) {
    std::cerr << "Runtime error: " << e.what() << std::endl ;
  } catch (std::exception e) {
    std::cerr << "Error: " << e.what() << std::endl ;
  }
}
