/*
 *
 * nucleo/core/URI.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/core/URI.H>

#include <iostream>

#include <stdio.h>

static void
split(std::string &s, const std::string splitters, std::string &part, bool keepsep, bool preserve) {
  std::string::size_type pos = s.find_first_of(splitters) ;
  if (pos!=std::string::npos) {    
    part.assign(s, 0, pos) ;
    s.erase(0, (keepsep ? pos : pos+1)) ;
  } else {
    if (!preserve) {
	 part.assign(s) ;
	 s = "" ;
    }
  }
}

namespace nucleo {

  // ----------------------------------------------------------------------

  void
  URI::load(const std::string &s) {
    std::string uri = s ;

    _scheme = "" ;
    _user = "" ;
    _password = "" ;
    _host = "" ;
    _port = 0 ;
    _path = "" ;
    _query = "" ;
    _fragment = "" ;
    _opaque = "" ;

    split(uri, ":", _scheme, false, true) ;

    _isAbsolute = (_scheme.length()>0) ;
    _isOpaque = false ;

    if (uri[0]=='/') {
	 if (uri[1]=='/') {
	   uri.erase(0,2) ;
	   std::string site ;
	   split(uri, "/?", site, true, false) ;
	   if (site.length()) {
		split(site, "@", _password, false, true) ;
		if (_password.length())
		  split(_password, ":", _user, false, false) ;
		split(site, ":", _host, false, false) ;
		_port = atoi(site.c_str()) ;
	   }
	 } 
    } else _isOpaque = _isAbsolute ;

    if (_isOpaque) {
	 // Was: split(uri, "#", _opaque, false, false) ;
	 split(uri, "?", _opaque, false, false) ;
	 split(uri, "#", _query, false, false) ;
    } else {
	 split(uri, "?", _path, false, false) ;
	 split(uri, "#", _query, false, false) ;
    }
 
    _fragment.assign(uri) ;
  }

  // ----------------------------------------------------------------------

  void
  URI::debug(std::ostream& out) const {
    out << "Kind     : " ;
    out << (_isAbsolute ? "ABSOLUTE " : "RELATIVE ") ;
    if (_isOpaque) out << "OPAQUE " ;
    out << std::endl ;

    if (_isAbsolute) 
	 out << "Scheme   : " << _scheme << std::endl ;

    if (!_isOpaque) {
	 out << "User     : " << _user << std::endl ;
	 out << "Password : " << _password << std::endl ;
	 out << "Host     : " << _host << std::endl ;
	 out << "Port     : " << _port << std::endl ;

	 out << "Path     : " << _path << std::endl ;
	 out << "Query    : " << _query << std::endl ;
    } else 
	 out << "OPAQUE   : " << _opaque << std::endl ;
    out << "Fragment : " << _fragment << std::endl ;

    out.flush() ;
  }

  // ----------------------------------------------------------------------

  std::string
  URI::encodeForQuery(const std::string &src, int flags) {
    static const char m[] = {
	 //                  1                   2                   3
	 //1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
	 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	 2,1,2,2,1,2,4,1,1,1,1,4,1,1,1,4,1,1,1,1,1,1,1,1,1,1,4,4,2,4,2,4,
	 4,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,1,
	 2,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,2,2,2,2,0,
	 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
	 0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
    };
    static const char* h = "0123456789ABCDEF";
    std::string result ;

    for( std::string::const_iterator i=src.begin();
	    i != src.end(); 
	    ++i ) {
	 char c = *i ;
	 if (m[(int)c]&flags)
	   result.append(&c,1) ;
	 else {
	   char tmp[3] ;
	   sprintf(tmp,"%%%c%c", h[c>>4], h[c&0x0F]) ;
	   result.append(tmp,3) ;
	 }
    }

    return result ;
  }

  // ----------------------------------------------------------------------

  inline static unsigned int
  hex2int(char *s) {
    unsigned int t = 0;
    for(int i=0; i<2; i++)
	 if(s[i]>='0' && s[i]<='9')
	   t = (t<<4)+(s[i]-'0');
	 else if(s[i]>='a' && s[i]<='f')
	   t = (t<<4)+(s[i]-'a'+10);
	 else if(s[i]>='A' && s[i]<='F')
	   t = (t<<4)+(s[i]-'A'+10);
	 else
	   break;
    return t ;
  }

  std::string
  URI::decodeFromQuery(const std::string &src) {
    std::string result ;

    for( std::string::const_iterator i=src.begin();
	    i != src.end(); 
	    ++i ) {
	 char c = *i ;   
	 if (c == '+') c = ' ' ;
	 else if (c == '%') {
	   char tmp[2] ;
	   ++i ; tmp[0] = *i ;
	   ++i ; tmp[1] = *i ;
	   c = (char)hex2int(tmp) ;
	 }
	 result.append(&c,1) ;
    }

    return result ;
  }

  // ----------------------------------------------------------------------

  bool
  URI::getQueryArg(const std::string &q, const std::string &key, std::string *value) {
    if (q=="") return false ;

    std::string::size_type from=0 ;
    do {
	 std::string::size_type b = q.find(key, from) ;
	 if (b==std::string::npos) return false ;

	 // found one candidate...
	 std::string::size_type next = b+key.length() ;
	 bool okLeft = (b==0 || q[b-1]=='&') ;
	 bool okRight = (next==q.length() || q[next]=='=' || q[next]=='&') ;
	 if (okLeft && okRight) {
	   // candidate seems good...
	   if (value) {
		*value = "" ;
		if (q[next]=='=' && next+1<q.length()) {
		  std::string::size_type e = q.find("&",next+1) ;
		  std::string::size_type n = (e==std::string::npos) ? (q.length()-next+1) : (e-next-1) ;
		  std::string result ;
		  result.assign(q, next+1, n) ;
		  *value = decodeFromQuery(result) ;
		}
	   }
	   return true ;
	 }
	 from = next ;
    } while (from<q.length()) ;

    return false ;
  }

  bool
  URI::getQueryArg(const std::string &q, const std::string &key, int *value) {
    std::string tmp ;
    if (!getQueryArg(q, key, &tmp)) return false ;
    *value = atoi(tmp.c_str()) ;
    return true ;
  }

  bool
  URI::getQueryArg(const std::string &q, const std::string &key, unsigned int *value) {
    std::string tmp ;
    if (!getQueryArg(q, key, &tmp)) return false ;
    *value = atoi(tmp.c_str()) ;
    return true ;
  }

  bool
  URI::getQueryArg(const std::string &q, const std::string &key, unsigned long *value) {
    std::string tmp ;
    if (!getQueryArg(q, key, &tmp)) return false ;
    *value = strtol(tmp.c_str(), (char**)NULL,0) ;
    return true ;
  }

  bool
  URI::getQueryArg(const std::string &q, const std::string &key, double *value) {
    std::string tmp ;
    if (!getQueryArg(q, key, &tmp)) return false ;
    *value = atof(tmp.c_str()) ;
    return true ;
  }

  bool
  URI::getQueryArg(const std::string &q, const std::string &key, float *value) {
    std::string tmp ;
    if (!getQueryArg(q, key, &tmp)) return false ;
    *value = atof(tmp.c_str()) ;
    return true ;
  }

  // ----------------------------------------------------------------------

}
