/*
 *
 * nucleo/image/source/ImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/source/ImageSource.H>

#ifdef __APPLE__
#include <nucleo/image/source/qtMovImageSource.H>
#include <nucleo/image/source/qtImgImageSource.H>
#include <nucleo/image/source/qtSgImageSource.H>
#endif

#ifdef __linux__
#include <nucleo/image/source/v4lImageSource.H>
#endif

#include <nucleo/image/source/nudpImageSource.H>
#include <nucleo/image/source/nudppImageSource.H>
#include <nucleo/image/source/nudpcImageSource.H>

#include <nucleo/image/source/serverpushImageSource.H>

#include <nucleo/image/source/imagefileImageSource.H>
#include <nucleo/image/source/nspImageSource.H>

#include <nucleo/image/source/noiseImageSource.H>

#include <nucleo/core/PluginManager.H>

#include <string>
#include <stdexcept>

namespace nucleo {

  // ------------------------------------------------------------------

  bool
  ImageSource::waitForImage(Image *img) {
    while (isActive()) {
	 if (getNextImage(img)) return true ;
	 ReactiveEngine::step() ;
    }
    return false ;
  }

  // ------------------------------------------------------------------

  bool
  ImageSource::getImage(const char *uri, Image *img, Image::Encoding enc) {
    ImageSource *source = ImageSource::create(uri, enc) ;
    source->start() ;
    bool gotTheImage = source->waitForImage(img) ;
    if (gotTheImage && img->dataIsLinked()) img->acquireData() ;
    delete source ;
    return gotTheImage ;
  }

  // ------------------------------------------------------------------

  ImageSource *
  ImageSource::create(const char *u, Image::Encoding e) {
    if (!u) throw std::runtime_error("Can't create an ImageSource from an empty URI...") ;

    URI uri(u) ;

    std::string scheme = uri.getScheme() ;
    if (scheme=="") scheme = "file" ;

    // -------------------------------------------------------

    ImageSourceFactory factory = 0 ;

#ifdef __APPLE__
    if (scheme == "videoin") return new qtSgImageSource(e, uri) ;
#endif
#ifdef __linux__
    if (scheme == "videoin") return new v4lImageSource(e, uri) ;
#endif
    if (scheme == "nudp") return new nudpImageSource(e, uri) ;
    if (scheme == "nudpp") return new nudppImageSource(e, uri) ;
    if (scheme == "nudpc") return new nudpcImageSource(e, uri) ;
    if (scheme=="http") return new serverpushImageSource(e, uri) ;
    if (scheme=="noise") return new noiseImageSource(e, uri) ;

    if (scheme=="file") {
	 std::string filename = uri.isOpaque() ? uri.getOpaque() : uri.getPath() ;
	 const char *extension = getExtension(filename.c_str()) ;
	 if (!extension)
	   throw std::runtime_error("createImageSource: unknown file type '"+filename+"'") ;
	 cistring iext = extension ;
	 if ((iext == ".vss") || (iext == ".nsp")) return new nspImageSource(e, uri) ;
	 if (iext == ".jpg") return new imagefileImageSource(e, uri, Image::JPEG) ;
	 if (iext == ".jpeg") return new imagefileImageSource(e, uri, Image::JPEG) ;
	 if (iext == ".png") return new imagefileImageSource(e, uri, Image::PNG) ;
	 if (iext == ".pam") return new imagefileImageSource(e, uri, Image::PAM) ;
#ifdef __APPLE__
	 try {
	   return new qtMovImageSource(e, uri) ; // or qtImgImageSource(e, uri) ?
	 } catch (...) {
	   std::cerr << "Failed to create a qtMovImageSource from " << filename << std::endl ;
	 }
#endif
	 std::string tag = std::string("extension=")+extension ;
	 try {
	   factory = (ImageSourceFactory) PluginManager::getSymbol("ImageSource::create", tag) ;
	   return (ImageSource*) ((*factory)(&uri, e)) ;
	 } catch (std::runtime_error err) {
	 }
    }

    factory = (ImageSourceFactory) PluginManager::getSymbol("ImageSource::create",
												std::string("scheme=")+scheme) ;
    return (ImageSource*) ((*factory)(&uri, e)) ;
  }

  // ------------------------------------------------------------------

}
