/*
 *
 * nucleo/image/source/nspImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/source/nspImageSource.H>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

namespace nucleo {

  // ------------------------------------------------------------------

  nspImageSource::nspImageSource(Image::Encoding encoding, const URI &uri) : ImageSource() {
    _encoding = encoding ;

    _filename = uri.isOpaque() ? uri.getOpaque() : uri.getPath() ;

    std::string query=uri.getQuery() ;

    _syncFromData = true ;
    _deltat = 0 ;

    double dArg ;
    if (URI::getQueryArg(query, "framerate", &dArg) && dArg) {
	 _syncFromData = false ;
	 _deltat = (unsigned long)(1000.0 / dArg) ;
    }

    if (URI::getQueryArg(query, "pause", &dArg)) {
	 _syncFromData = false ;
	 _deltat = (unsigned long)(1000.0 * dArg) ;
    }

    // std::cerr << "nspImageSource::nspImageSource: deltat = " << _deltat << std::endl ;

    _state = CLOSED ;
  }

  bool
  nspImageSource::start(void) {
    if (_state!=CLOSED) return false ;

    _msg.reset() ;
    _fd = open(_filename.c_str(), O_RDONLY) ;
    if (_fd==-1) {
	 std::cerr << "nspImageSource: no such file (" << _filename << ")" << std::endl ;
	 return false ;
    }
    _state = OPEN ;
    _chrono.start() ;
    _tk = TimeKeeper::create(_deltat*TimeKeeper::millisecond, true) ;
    // std::cerr << "nspImageSource " << this << " subscribes to TimeKeeper " << _tk << std::endl ;
    subscribeTo(_tk) ;
    return true ;
  }

  void
  nspImageSource::react(Observable*) {
    // std::cerr << "nspImageSource::react" << std::endl ;

    // We need an image! But we can't use parseFromStream cause it
    // would read the entire file at once...
    while (_msg.parseData()!=HttpMessage::COMPLETE) {
	 int nbb = _msg.feedFromStream(_fd) ;
	 // std::cerr << nbb << " bytes read in nspImageSource::react" << std::endl ;
	 if (nbb<1) {
	   if (_msg.completeData()==HttpMessage::COMPLETE) {
		break ;
	   }
	   stop() ;
	   notifyObservers() ; // make sure the observers will know about our new state
	   return ;
	 }
    }

    _state = READY ;
    if (!_pendingNotifications) notifyObservers() ;
  }

  bool
  nspImageSource::getNextImage(Image *img) {
    // std::cerr << "nspImageSource::getNextImage" << std::flush ;

    if (_state!=READY) return false ; 

    if (_syncFromData) {
	 std::string framerate ;
	 if (_msg.getHeader("nucleo-framerate", framerate)) {
	   double fr = atof(framerate.c_str()) ;
	   if (fr) {
		unsigned long dt = (unsigned long)(1000.0/fr) ;
		// std::cerr << "New deltat from data: " << dt << " (" << fr << ")" << std::endl ;
		_tk->arm(dt*TimeKeeper::millisecond, true) ;
	   }
	 }
    }

    std::string smimetype, swidth, sheight ;
    unsigned int width=0, height=0 ;
    Image::Encoding encoding = Image::JPEG ;
    if (_msg.getHeader("nucleo-image-width", swidth))
	 width = atoi(swidth.c_str()) ;
    if (_msg.getHeader("nucleo-image-height", sheight))
	 height = atoi(sheight.c_str()) ;
    if (_msg.getHeader("content-type", smimetype))
	 encoding = Image::getEncodingByMimeType(smimetype) ;
    if (encoding==Image::OPAQUE) encoding = Image::JPEG ;

#if 0
    std::cerr << width << "x" << height << " " << Image::getEncodingName(encoding) 
		    << " (" << smimetype << ")"
		    << std::endl ;
#endif

    std::string const &body = _msg.body() ;
    img->setEncoding(encoding) ;
    img->setDims(width, height) ;
    img->setData((unsigned char *)body.c_str(), body.length(), Image::NONE) ;

    if (!convertImage(img, _encoding)) return false ;

    // If no conversion was necessary, we must copy the data since it
    // will be overwritten by the _msg.next() call below
    if (img->getData()==(unsigned char *)body.c_str())
	 img->acquireData() ;

    _chrono.tick() ;

    _msg.next() ;	 
    _state = OPEN ;

    return true ;
  }

  bool
  nspImageSource::stop(void) {
    // std::cerr << "nspImageSource::stop" << std::endl ;

    if (_state==CLOSED) return false ;
    _chrono.stop() ;
    close(_fd) ;
    _state = CLOSED ;
    unsubscribeFrom(_tk) ;
    delete _tk ;
    return true ;
  }

  // ------------------------------------------------------------------

}
