/*
 *
 * nucleo/image/source/nudpcImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/source/nudpcImageSource.H>

#include <nucleo/utils/FileUtils.H>

#include <netdb.h>
#include <stdio.h>
#include <unistd.h>

#include <stdexcept>

#define DEBUG_LEVEL 0

namespace nucleo {

  // ------------------------------------------------------------------

  nudpcImageSource::nudpcImageSource(Image::Encoding target_encoding, const URI &uri) : ImageSource() {
#if DEBUG_LEVEL>=1
    std::cerr << "nudpcImageSource::nudpcImageSource" << std::endl ;
#endif
 
    _hostname = uri.getHost() ;
    if (_hostname=="") _hostname="localhost" ;
    _port = uri.getPort() ;
    if (!_port) _port = 5555 ;

    std::string query = uri.getQuery() ;
   
    _target_encoding = target_encoding==Image::PREFERRED ? Image::JPEG : target_encoding ;

    _request = "/nudp" ;
    std::string tmp = uri.getPath() ;
    if (tmp!="") _request.append(tmp) ; else _request.append("/video") ;
    _request.append("?") ;
    if (query!="") {
	 _request.append(query) ;
	 _request.append("&") ;
    }

    _state = CLOSED ;
  }

  bool
  nudpcImageSource::start(void) {
#if DEBUG_LEVEL>=1
    std::cerr << "nudpcImageSource::start" << std::endl ;
#endif

    if (_state!=CLOSED) return false ;

    try {
	 _connection = new TcpConnection(_hostname, _port) ;
    } catch (std::runtime_error e) {
	 std::cerr << "nudpcImageSource (start): " << e.what() << std::endl ;
	 return false ;
    } catch (...) {
	 return false ;
    }

    _udp = new UdpReceiver() ;

    // --- Get the local hostname and _udp port number -------------------

    char hostname[50] ;
    gethostname(hostname, 50) ;
    char udpOptions[255] ;

    hostent* local_hostent=gethostbyname(hostname) ;
    sprintf(udpOptions, "nudp=%d.%d.%d.%d%%3A%d", (unsigned char)(local_hostent->h_addr[0]), (unsigned char)(local_hostent->h_addr[1]), (unsigned char)(local_hostent->h_addr[2]), (unsigned char)(local_hostent->h_addr[3]), ntohs(_udp->getPortNumber())) ;

    // ---------------------------------------------------------------------

    std::string request = "GET " ;
    request.append(_request) ;
    request.append(udpOptions) ;
    request.append(" HTTP/1.0") ;
    request.append(oneCRLF) ;
    request.append(oneCRLF) ;

    try {
	 _connection->send((const char*) request.c_str(), request.length()) ; 
    } catch (std::runtime_error e) {
	 std::cerr << "nudpcImageSource (start, 2): " << e.what() << std::endl ;
	 delete _connection ;
	 return false ;
    } catch (...) {
	 delete _connection ;
	 return false ;
    }

    subscribeTo(_connection) ;
    subscribeTo(_udp) ;

    _state = OPEN ;
    _chrono.start() ;
    return true ;
  }

  void
  nudpcImageSource::react(Observable *obs) {
#if DEBUG_LEVEL>=1
    std::cerr << "nudpcImageSource::react" << std::endl ;
#endif

    if (obs==_udp) {
	 if (_udp->receive(&_data, &_dataSize)) {
	   _state = READY ;
	   if (!_pendingNotifications) notifyObservers() ;
	 }
    }

    if (obs==_connection) {
	 int fd = _connection->getFd() ;
	 HttpMessage msg ; 
	 msg.parseFromStream(fd) ;
	 // maybe we could do something with msg...
	 stop() ;
	 notifyObservers() ;
    }
  }

  bool
  nudpcImageSource::getNextImage(Image *img) {
#if DEBUG_LEVEL>=1
    std::cerr << "nudpcImageSource::getNextImage" << std::endl ;
#endif
    if (_state!=READY) return false ; 

    img->setEncoding(Image::JPEG) ;
    img->setData(_data, _dataSize, Image::DELETE) ;
    _state = OPEN ;
    if (!convertImage(img, _target_encoding)) return false ;
    _chrono.tick() ;
    return true ;
  }

  bool
  nudpcImageSource::stop(void) {
#if DEBUG_LEVEL>=1
    std::cerr << "nudpcImageSource::stop" << std::endl ;
#endif

    if (_state==CLOSED) return false ;

    _chrono.stop() ;

    unsubscribeFrom(_udp) ;
    delete _udp ;

    unsubscribeFrom(_connection) ;
    delete _connection ;

    _state = CLOSED ;
    return true ;
  }

  // ------------------------------------------------------------------

}
