/*
 *
 * nucleo/image/source/serverpushImageSource.cxx --
 *
 * Copyright (C) Nicolas Roussel
 *
 * See the file LICENSE for information on usage and redistribution of
 * this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 */

#include <nucleo/config.H>

#include <nucleo/image/source/serverpushImageSource.H>

#include <nucleo/utils/FileUtils.H>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <stdexcept>
#include <sstream>

namespace nucleo {

  // ------------------------------------------------------------------

  serverpushImageSource::serverpushImageSource(Image::Encoding encoding, const URI &uri) : ImageSource() {
    _encoding = encoding ; 

    _hostname = uri.getHost() ;
    if (_hostname=="") _hostname="localhost" ;
    _port = uri.getPort() ;
    if (!_port) _port = 80 ;

    std::string resource = uri.getPath() ;
    if (resource=="") resource = "/" ;
    std::string query = uri.getQuery() ;
    if (query!="") resource = resource + "?" + query ;
    std::stringstream request ;
    request << "GET " << resource << " HTTP/1.1" << oneCRLF
		  << "User-Agent: Mozilla/1.1 nucleo/" << PACKAGE_VERSION << oneCRLF
		  << "Accept: image/jpeg" << oneCRLF
#if 0
		  << "Connection: keep-alive" << oneCRLF
		  << "Connection: close" << oneCRLF
#endif
		  << "Host: " << _hostname << oneCRLF
		  << oneCRLF ;
    _request = request.str() ;

    _state = CLOSED ;
    _connection = 0 ;
  }

  bool
  serverpushImageSource::start(void) {
    if (_state!=CLOSED) return false ;

    _msg.reset() ;
    try {
	 _connection = new TcpConnection(_hostname, _port) ;
    } catch (std::runtime_error e) {
	 std::cerr << "serverpushImageSource: " << e.what() << std::endl ;
	 return false ;
    } catch (...) {
	 return false ;
    }

    subscribeTo(_connection) ;
	
    try {
	 _connection->send((const char*) _request.c_str(), _request.length()) ;
    } catch (std::runtime_error e) {
	 std::cerr << "serverpushImageSource: " << e.what() << std::endl ;
	 delete _connection ;
	 return false;
    } catch (...) {
	 return false ;
    }
	
    _state = OPEN ;
    _chrono.start() ;
    return true ;
  }

  void
  serverpushImageSource::react(Observable *obs) {
    switch(_state) {
    case OPEN:
	 if (obs==_connection) {
	   int fd = _connection->getFd() ;
	   if (_msg.feedFromStream(fd)<1) _state = BUFFERED ;
	   notifyObservers() ;
	 }
	 break ;
    case BUFFERED:
	 if (!_msg.dataWaiting()) stop() ;
	 notifyObservers() ;
	 break ;
    default:
	 break ;
    }
  }

  bool
  serverpushImageSource::getNextImage(Image *img) {
    if (_state==CLOSED) return false ; 

    HttpMessage::State s = _msg.parseData() ;
    if (s==HttpMessage::NEED_BODY && _state==BUFFERED) s=_msg.completeData() ;

    if (s!=HttpMessage::COMPLETE) {
	 if (_state==BUFFERED) {
	   stop() ;
	   notifyObservers() ;
	 }
	 return false ;
    }

    std::string smimetype, swidth, sheight ;
    unsigned int width=0, height=0 ;
    Image::Encoding encoding = Image::JPEG ;
    if (_msg.getHeader("nucleo-image-width", swidth))
	 width = atoi(swidth.c_str()) ;
    if (_msg.getHeader("nucleo-image-height", sheight))
	 height = atoi(sheight.c_str()) ;
    if (_msg.getHeader("content-type", smimetype))
	 encoding = Image::getEncodingByMimeType(smimetype) ;
    if (encoding==Image::OPAQUE) encoding = Image::JPEG ;

#if 0
    std::cerr << width << "x" << height << " " << Image::getEncodingName(encoding) 
		    << " (" << smimetype << ")"
		    << std::endl ;
#endif

    std::string const &body = _msg.body() ;
    img->setEncoding(encoding) ;
    img->setDims(width, height) ;
    img->setData((unsigned char *)body.c_str(), body.length(), Image::NONE) ;
    convertImage(img, _encoding) ;

    // If no conversion was necessary, we must copy the data since it
    // will be overwritten by the _msg.next() call below
    if (img->getData()==(unsigned char *)body.c_str())
	 img->acquireData() ;

    _chrono.tick() ;
    _msg.next() ;	 
    return true ;
  }

  bool
  serverpushImageSource::stop(void) {
    if (_state!=CLOSED) {
	 _chrono.stop() ;
	 unsubscribeFrom(_connection) ;
	 delete _connection ;
	 _state = CLOSED ;
	 return true ;
    }
    return false ;
  }

  // ------------------------------------------------------------------

}
